from pydantic_core import Url

from kptncook.api import KptnCookClient


def test_client_to_url():
    base_url = Url(
        "https://mobile.kptncook.com"
    )  # make sure urljoin works with pydantic URLs
    client = KptnCookClient(base_url=base_url)
    assert client.to_url("/recipes") == "https://mobile.kptncook.com/recipes"
    assert client.to_url("recipes") == "https://mobile.kptncook.com/recipes"



"""
Export a single recipe to Paprika App

file format:
    1. export recipe to json
    2. compress file as gz: naming convention: a_recipe_name.paprikarecipe (Singular)
    3. zip this file as some_recipes.paprikarecipes (Plural!)
"""
import base64
import glob
import gzip
import os
import re
import secrets
import shutil
import tempfile
import zipfile
from datetime import datetime
from pathlib import Path
from typing import Any

import httpx
from jinja2 import Template
from unidecode import unidecode

from kptncook.config import settings
from kptncook.models import Image, Recipe

PAPRIKA_RECIPE_TEMPLATE = """{
   "uid":"{{recipe.id.oid}}",
   "name":"{{recipe.localized_title.de}}",
   "directions": "{% for step in recipe.steps %}{{step.title.de}}\\n{% endfor %}",
   "servings":"2",
   "rating":0,
   "difficulty":"",
   "ingredients":"{% for ingredient in recipe.ingredients %}{% if ingredient.quantity %}{{'{0:g}'.format(ingredient.quantity) }}{% endif %} {{ingredient.measure|default('',true)}} {{ingredient.ingredient.uncountable_title.de|default('',true)}}\n{% endfor %}",
   "notes":"",
   "created":"{{dtnow}}",
   "image_url":null,
   "cook_time":"{{recipe.cooking_time|default('',true)}}",
   "prep_time":"{{recipe.preparation_time|default('',true)}}",
   "source":"Kptncook",
   "source_url":"",
   "hash" : "{{hash}}",
   "photo_hash":null,
   "photos":[],
   "photo": "{{cover_filename}}",
   "nutritional_info":"{% for nutrient, amount in recipe.recipe_nutrition %}{{nutrient}}: {{amount}}\\n{% endfor %}",
   "photo_data":"{{cover_img}}",
   "photo_large":null,
   "categories":["Kptncook"]
}
"""  # noqa: E501


class GeneratedData:
    def __init__(
        self, cover_filename: str | None, cover_img: str | None, dtnow: str, hash_: str
    ):
        self.cover_filename = cover_filename
        self.cover_img = cover_img
        self.dtnow = dtnow
        self.hash = hash_


class PaprikaExporter:
    def export(self, recipes: list[Recipe]) -> str:
        export_data = self.get_export_data(recipes=recipes)
        filename = self.get_export_filename(export_data=export_data, recipes=recipes)
        tmp_dir = tempfile.mkdtemp()
        filename_full_path = self.save_recipes(
            export_data=export_data, directory=tmp_dir, filename=filename
        )
        self.move_to_target_dir(
            source=filename_full_path, target=os.path.join(str(Path.cwd()), filename)
        )
        return filename

    def get_export_filename(
        self, export_data: dict[str, str], recipes: list[Recipe]
    ) -> str:
        if len(export_data) == 1:
            return (
                self.asciify_string(s=recipes[0].localized_title.de) + ".paprikarecipes"
            )
        else:
            return "allrecipes.paprikarecipes"

    def get_generated_data(self, recipe: Recipe) -> GeneratedData:
        """Just to make testing easier and only have one method to mock in tests."""
        cover_filename, cover_img = self.get_cover_img_as_base64_string(recipe=recipe)
        dtnow = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        hash = secrets.token_hex(32)
        return GeneratedData(cover_filename, cover_img, dtnow, hash)

    def get_export_data(self, recipes: list[Recipe]) -> dict[str, str]:
        template = Template(PAPRIKA_RECIPE_TEMPLATE, trim_blocks=True)
        export_data = dict()
        for recipe in recipes:
            generated = self.get_generated_data(recipe=recipe)
            recipe_as_json = template.render(
                recipe=recipe,
                dtnow=generated.dtnow,
                cover_filename=generated.cover_filename,
                hash=generated.hash,
                cover_img=generated.cover_img,
            )
            export_data[str(recipe.id.oid)] = recipe_as_json
        return export_data

    def move_to_target_dir(self, source: str, target: str) -> str:
        return shutil.move(source, target)

    def asciify_string(self, s) -> str:
        s = unidecode(s)
        s = re.sub(r"[^\w\s]", "_", s)
        s = re.sub(r"\s+", "_", s)
        return s

    def save_recipes(
        self, export_data: dict[str, Any], filename: str, directory: str
    ) -> str:
        for id, recipe_as_json in export_data.items():
            recipe_as_gz = os.path.join(directory, "recipe_" + id + ".paprikarecipe")
            with gzip.open(recipe_as_gz, "wb") as f:
                f.write(recipe_as_json.encode("utf-8"))
        filename_full_path = os.path.join(directory, filename)
        with zipfile.ZipFile(
            filename_full_path, "w", compression=zipfile.ZIP_DEFLATED, allowZip64=True
        ) as zip_file:
            gz_files = glob.glob(os.path.join(directory, "*.paprikarecipe"))
            print(gz_files)
            for gz_file in gz_files:
                zip_file.write(gz_file, arcname=os.path.basename(gz_file))
        return filename_full_path

    def get_cover_img_as_base64_string(
        self, recipe: Recipe
    ) -> tuple[str | None, str | None]:
        cover = self.get_cover(image_list=recipe.image_list)
        if cover is None:
            raise ValueError("No cover image found")
        cover_url = recipe.get_image_url(api_key=settings.kptncook_api_key)
        if not isinstance(cover_url, str):
            raise ValueError("Cover URL must be a string")
        try:
            response = httpx.get(cover_url)
            response.raise_for_status()
        except httpx.HTTPStatusError as exc:
            if exc.response.status_code == 404:
                print(
                    f'Cover image for "{recipe.localized_title.de}" not found online any more.'
                )
            else:
                print(
                    f"While trying to fetch the cover img a HTTP error occurred: {exc.response.status_code}: {exc}"
                )
            return None, None
        return cover.name, base64.b64encode(response.content).decode("utf-8")

    def get_cover(self, image_list: list[Image]) -> Image | None:
        if not isinstance(image_list, list):
            raise ValueError("Parameter image_list must be a list")
        try:
            [cover] = [i for i in image_list if i.type == "cover"]
        except ValueError:
            return None
        return cover



from kptncook.mealie import kptncook_to_mealie
from kptncook.models import Recipe


def test_parse_full_recipe(full_recipe):
    kc_recipe = Recipe.model_validate(full_recipe)
    mealie_recipe = kptncook_to_mealie(kc_recipe)
    assert mealie_recipe.name == kc_recipe.localized_title.de
    assert mealie_recipe.tags is not None
    # assert "kptncook" in mealie_recipe.tags  # tags do not work atm
    assert mealie_recipe.extras is not None
    assert "kptncook_id" in mealie_recipe.extras
    assert mealie_recipe.extras["source"] == "kptncook"



import json
from pathlib import Path

import pytest


@pytest.fixture
def full_recipe():
    example_path = Path(__file__).parent / "fixtures" / "kptncook_example.json"
    with example_path.open("r") as f:
        example = json.load(f)
    return example


@pytest.fixture
def minimal():
    return {
        "_id": {"$oid": "5e5390e2740000cdf1381c64"},
        "localizedTitle": {"de": "Minimal Recipe"},
        "country": "us/de/ww",
        "authorComment": {"de": "Dies ist ein Kommentar"},
        "preparationTime": 20,
        "recipeNutrition": {
            "calories": 100,
            "fat": 10,
            "carbohydrate": 20,
            "protein": 30,
        },
        "steps": [
            {
                "title": {"de": "Alles parat?"},
                "ingredients": [],
                "image": {
                    "name": "REZ_6666_11.jpg",
                    "url": "https://d2am1qai33sroc.cloudfront.net/image/63652er8d4b00007500b0c51d",
                    "type": "step",
                },
            }
        ],
        "imageList": [
            {
                "name": "REZ_1837_Cover.jpg",
                "type": "cover",
                "url": "https://d2am1qai33sroc.cloudfront.net/image/f6813160-68a5-420f-8c77-be9dcc2fa91b",
            },
        ],
        "ingredients": [],
    }



"""
All the domain models for kptncook live here.
"""

from pydantic import BaseModel, Field, model_validator


def to_camel(string: str) -> str:
    """Convert snake_case to camelCase."""
    words = []
    for num, word in enumerate(string.split("_")):
        if num > 0:
            word = word.capitalize()
        words.append(word)
    return "".join(words)


class LocalizedString(BaseModel):
    en: str | None = None
    de: str | None = None
    es: str | None = None
    fr: str | None = None
    pt: str | None = None


class Nutrition(BaseModel):
    calories: int
    protein: int
    fat: int
    carbohydrate: int


class Image(BaseModel):
    name: str
    type: str | None = None
    url: str

    def get_image_with_api_key_url(self, api_key: str) -> "Image":
        url_with_key = f"{self.url}?kptnkey={api_key}"
        kwargs = self.model_dump() | {"url": url_with_key}
        return Image(**kwargs)


class IngredientDetails(BaseModel):
    typ: str
    localized_title: LocalizedString
    number_title: LocalizedString
    uncountable_title: LocalizedString | None = None
    category: str

    @model_validator(mode="before")
    def fix_json_errors(cls, values):
        if "uncountableTitle" not in values or values["uncountableTitle"] is None:
            if "numberTitle" in values:
                values["uncountableTitle"] = values["numberTitle"]
        return values

    class Config:
        alias_generator = to_camel


class Ingredient(BaseModel):
    quantity: float | None = None
    measure: str | None = None
    ingredient: IngredientDetails


class RecipeId(BaseModel):
    oid: str = Field(..., alias="$oid")


class RecipeStep(BaseModel):
    title: LocalizedString
    image: Image


class Recipe(BaseModel):
    id: RecipeId = Field(..., alias="_id")
    localized_title: LocalizedString
    country: str
    author_comment: LocalizedString
    preparation_time: int
    cooking_time: int | None = None
    recipe_nutrition: Nutrition
    steps: list[RecipeStep] = Field(..., alias="steps")
    image_list: list[Image]
    ingredients: list[Ingredient]

    class Config:
        alias_generator = to_camel

    def get_image_url(self, api_key: str) -> str | None:
        try:
            [image] = [i for i in self.image_list if i.type == "cover"]
        except ValueError:
            return None
        image_url = f"{image.url}?kptnkey={api_key}"
        return image_url



import datetime
import io
import json
import logging
from getpass import getpass
from pathlib import Path
from typing import Any

import httpx
from pydantic import UUID4, BaseModel, Field, ValidationError, parse_obj_as

from .config import settings
from .models import Image
from .models import Recipe as KptnCookRecipe

logger = logging.getLogger(__name__)


class NameIsIdModel(BaseModel):
    name: str

    def __hash__(self):
        """Hash on name to be able to have sets of models."""
        return hash(self.name)

    def __eq__(self, other):
        """Compare on name to be able to subtract sets of models."""
        return self.name == other.name

    class Config:
        frozen = True  # make it hashable


class RecipeTag(NameIsIdModel):
    slug: str | None = None
    name: str
    group_id: UUID4 | None = Field(None, alias="groupId")
    id: UUID4 | None = None


class RecipeCategory(RecipeTag):
    pass


class RecipeTool(RecipeTag):
    on_hand: bool = False


class UnitFoodBase(NameIsIdModel):
    id: UUID4 | None = None
    name: str
    description: str = ""


class RecipeFood(UnitFoodBase):
    ...


class RecipeUnit(UnitFoodBase):
    id: UUID4 | None = None
    fraction: bool = True
    abbreviation: str = ""


class RecipeIngredient(BaseModel):
    title: str | None = None
    note: str | None = None
    unit: RecipeUnit | None
    food: RecipeUnit | None
    disable_amount: bool = True
    quantity: float | None = 1


class RecipeSummary(BaseModel):
    id: UUID4 | None = None

    user_id: UUID4 | None = Field(None, alias="userId")
    group_id: UUID4 | None = Field(None, alias="groupId")

    name: str | None = None
    slug: str = ""
    image: Any | None = None
    recipe_yield: str | None = None

    total_time: str | None = None
    prep_time: str | int | None = None
    cook_time: str | int | None = None
    perform_time: str | None = None

    description: str | None = ""
    recipe_category: list[str] | None = []
    tags: list[RecipeTag] | None = []
    # tags: list[RecipeTag | str] | None = []
    tools: list[RecipeTool] = []
    rating: int | None = None
    org_url: str | None = Field(None, alias="orgURL")

    recipe_ingredient: list[RecipeIngredient] | None = []

    date_added: datetime.date | None = None
    date_updated: datetime.datetime | None = None


class RecipeStep(BaseModel):
    title: str | None = ""
    text: str
    ingredientReferences: list[Any] = []
    image: Image | None = None


class Nutrition(BaseModel):
    calories: str | int | None = None
    fat_content: str | None = None
    protein_content: str | None = None
    carbohydrate_content: str | None = None
    fiber_content: str | None = None
    sodium_content: str | None = None
    sugar_content: str | None = None


class RecipeSettings(BaseModel):
    public: bool = True
    show_nutrition: bool = True
    show_assets: bool = False
    landscape_view: bool = False
    disable_comments: bool = False
    disable_amount: bool = False
    locked: bool = False


class RecipeAsset(BaseModel):
    name: str
    icon: str
    file_name: str | None = None


class RecipeNote(BaseModel):
    title: str
    text: str | None = None


class Recipe(RecipeSummary):
    recipe_ingredient: list[RecipeIngredient] = []
    recipe_instructions: list[RecipeStep] | None = []
    nutrition: Nutrition | None = None

    # Mealie Specific
    settings: RecipeSettings | None = RecipeSettings()
    assets: list[RecipeAsset] | None = []
    notes: list[RecipeNote] | None = []
    extras: dict = {}


class RecipeWithImage(Recipe):
    image_url: str


class MealieApiClient:
    def __init__(self, base_url):
        self.base_url = base_url
        self.headers = {}

    @property
    def logged_in(self):
        return "access_token" in self.headers

    def to_url(self, path):
        return f"{self.base_url}{path}"

    def __getattr__(self, name):
        """
        Return proxy for httpx, joining base_url with path and
        providing authentication headers automatically.
        """

        def proxy(path, **kwargs):
            url = self.to_url(path)
            set_headers = kwargs.get("headers", {})
            kwargs["headers"] = set_headers | self.headers
            with httpx.Client() as client:
                response = getattr(client, name)(url, **kwargs)
            return response

        return proxy

    def fetch_api_token(self, username, password):
        login_data = {"username": username, "password": password}
        r = self.post("/auth/token", data=login_data, timeout=60)
        r.raise_for_status()
        return r.json()["access_token"]

    def login(self, username: str = "admin", password: str = ""):
        if password == "":
            password = getpass()
        access_token = self.fetch_api_token(username, password)
        self.headers = {"authorization": f"Bearer {access_token}"}

    def upload_asset(self, recipe_slug, image: Image):
        # download image
        r = httpx.get(image.url)
        r.raise_for_status()

        download = io.BytesIO(r.content)

        # upload to mealie
        name = Path(image.name)
        extension = name.suffix.lstrip(".")
        stem = name.stem
        data = {
            "name": stem,
            "icon": "mdi-file-image",
            "extension": extension,
        }
        r = self.post(
            f"/recipes/{recipe_slug}/assets", data=data, files={"file": download}
        )
        r.raise_for_status()

        return r.json()

    @staticmethod
    def _build_recipestep_text(recipe_uuid, text, image_name):
        return f'{text} <img src="/api/media/recipes/{recipe_uuid}/assets/{image_name}" height="100%" width="100%"/>'

    def enrich_recipe_with_step_images(self, recipe):
        assets = []
        for instruction in recipe.recipe_instructions:
            asset_properties = self.upload_asset(recipe.slug, instruction.image)
            uploaded_image_name = asset_properties["fileName"]
            instruction.text = self._build_recipestep_text(
                recipe.id, instruction.text, uploaded_image_name
            )
            assets.append(RecipeAsset(name=asset_properties["name"], icon=asset_properties["icon"], file_name=asset_properties["fileName"]))
        recipe.assets = assets
        return recipe

    def _post_recipe_trunk_and_get_slug(self, recipe_name):
        data = {"name": recipe_name}
        r = self.post("/recipes", data=json.dumps(data))
        r.raise_for_status()
        slug = r.json()
        return slug

    def _scrape_image_for_recipe(self, recipe, slug):
        json_image_url = json.dumps({"url": recipe.image_url})
        scrape_image_path = f"/recipes/{slug}/image"
        r = self.post(scrape_image_path, data=json_image_url)
        r.raise_for_status()

    def _update_user_and_group_id(self, recipe, slug):
        recipe_detail_path = f"/recipes/{slug}"
        r = self.get(recipe_detail_path)
        r.raise_for_status()
        recipe_details = r.json()
        update_attributes = ["id", "userId", "groupId"]
        updated_details = {k: recipe_details[k] for k in update_attributes}
        recipe = RecipeWithImage(**(recipe.dict() | updated_details))
        return recipe

    def _get_page(self, endpoint_name, page_num, per_page=50):
        r = self.get(f"/{endpoint_name}?page={page_num}&perPage={per_page}")
        r.raise_for_status()
        return r.json()

    def _get_all_items(self, endpoint_name):
        all_items = []
        response_data = self._get_page(endpoint_name, 1)
        all_items.extend(response_data["items"])

        # 1 was already fetched, start page_num at 2 and add 1 to the
        # number of total pages, because we start counting at 1 instead of 0
        for page_num in range(2, response_data["total_pages"] + 1):
            response_data = self._get_page(endpoint_name, page_num)
            all_items.extend(response_data["items"])

        return all_items

    def _create_item(self, endpoint_name, item):
        r = self.post(f"/{endpoint_name}", data=item.json())
        r.raise_for_status()
        return r.json()

    def _create_item_name_to_item_lookup(self, endpoint_name, model_class, items):
        existing_items = parse_obj_as(
            set[model_class], self._get_all_items(endpoint_name)
        )
        items_to_create = items - existing_items
        for item in items_to_create:
            existing_items.add(model_class(**self._create_item(endpoint_name, item)))
        return {i.name: i for i in existing_items}

    def _update_item_ids(self, recipe, endpoint_name, model_class, attr_name):
        items = {
            getattr(ig, attr_name)
            for ig in recipe.recipe_ingredient
            if getattr(ig, attr_name) is not None
        }
        if len(items) == 0:
            # return early if there's nothing to do
            return recipe

        name_to_item_with_id = self._create_item_name_to_item_lookup(
            endpoint_name, model_class, items
        )
        for ingredient in recipe.recipe_ingredient:
            if getattr(ingredient, attr_name) is not None:
                setattr(
                    ingredient,
                    attr_name,
                    name_to_item_with_id[getattr(ingredient, attr_name).name],
                )
        return recipe

    def _update_tag_ids(self, recipe):
        recipe_tags = {tag for tag in recipe.tags}
        if len(recipe_tags) == 0:
            # return early if there's nothing to do
            return recipe

        name_to_tag_with_id = self._create_item_name_to_item_lookup(
            "organizers/tags", RecipeTag, recipe_tags
        )
        recipe.tags = [name_to_tag_with_id[tag.name] for tag in recipe_tags]
        return recipe

    def _update_recipe(self, recipe, slug):
        recipe_detail_path = f"/recipes/{slug}"
        r = self.put(recipe_detail_path, data=recipe.json())
        r.raise_for_status()
        return Recipe.model_validate(r.json())

    def create_recipe(self, recipe):
        slug = self._post_recipe_trunk_and_get_slug(recipe.name)
        print(slug)
        recipe.slug = slug
        self._scrape_image_for_recipe(recipe, slug)
        recipe = self._update_user_and_group_id(recipe, slug)
        recipe = self._update_item_ids(recipe, "units", RecipeUnit, "unit")
        recipe = self._update_item_ids(recipe, "foods", RecipeFood, "food")
        recipe = self._update_tag_ids(recipe)
        recipe = self.enrich_recipe_with_step_images(recipe)
        return self._update_recipe(recipe, slug)

    @staticmethod
    def validate_recipes(recipes):
        validated_recipes = []
        for recipe in recipes:
            try:
                validated_recipe = Recipe.model_validate(recipe)
                validated_recipes.append(validated_recipe)
            except ValidationError as e:
                logger.error(
                    "Could not parse recipe {recipe_id}".format(recipe_id=recipe["id"])
                )
                logger.exception(e)
                continue
        return validated_recipes

    def get_all_recipes(self):
        all_recipes = []

        r = self.get("/recipes?page=1&perPage=50")
        r.raise_for_status()
        all_recipes.extend(self.validate_recipes(r.json()["items"]))

        page = 2
        while page <= r.json()["total_pages"]:
            r = self.get(f"/recipes?page={page}&perPage=50")
            r.raise_for_status()
            all_recipes.extend(self.validate_recipes(r.json()["items"]))
            page += 1

        return all_recipes

    def delete_via_slug(self, slug):
        r = self.delete(f"/recipes/{slug}")
        r.raise_for_status()
        return r.json()

    def get_via_slug(self, slug):
        r = self.get(f"/recipes/{slug}")
        r.raise_for_status()
        return Recipe.model_validate(r.json())


def kptncook_to_mealie_ingredients(kptncook_ingredients):
    mealie_ingredients = []
    for ingredient in kptncook_ingredients:
        title = ingredient.ingredient.localized_title.de
        note = None
        if "," in title:
            title, note, *parts = (p.strip() for p in title.split(","))
        food = {"name": title}
        quantity = ingredient.quantity
        measure = None
        if hasattr(ingredient, "measure"):
            if ingredient.measure is not None:
                measure = {"name": ingredient.measure}
        mealie_ingredient = RecipeIngredient(
            title=title, quantity=quantity, unit=measure, note=note, food=food
        )
        mealie_ingredients.append(mealie_ingredient)
    return mealie_ingredients


def kptncook_to_mealie_steps(steps, api_key):
    mealie_instructions = []
    for step in steps:
        image = step.image.get_image_with_api_key_url(api_key)
        mealie_instructions.append(
            RecipeStep(title=None, text=step.title.de, image=image)
        )
    return mealie_instructions


def kptncook_to_mealie(
    kcin: KptnCookRecipe, api_key: str = settings.kptncook_api_key
) -> RecipeWithImage:
    kwargs = {
        "name": kcin.localized_title.de,
        "notes": [
            RecipeNote(title="author comment", text=kcin.author_comment.de),
        ],
        "nutrition": Nutrition(
            calories=kcin.recipe_nutrition.calories,
            proteinContent=kcin.recipe_nutrition.protein,
            fatContent=kcin.recipe_nutrition.fat,
            carbohydrateContent=kcin.recipe_nutrition.carbohydrate,
        ),  # type: ignore
        "prep_time": kcin.preparation_time,
        "cook_time": kcin.cooking_time,
        "recipe_yield": "1 Portionen",
        "recipe_instructions": kptncook_to_mealie_steps(kcin.steps, api_key),
        "recipe_ingredient": kptncook_to_mealie_ingredients(kcin.ingredients),
        "image_url": kcin.get_image_url(api_key),
        "tags": [RecipeTag.model_validate({"name": "kptncook", "group_id": None})],
        "extras": {"kptncook_id": kcin.id.oid, "source": "kptncook"},
    }
    print("kwargs: ", kwargs)
    return RecipeWithImage(**kwargs)



import json
from datetime import date

from kptncook.repositories import RecipeInDb, RecipeRepository


def test_no_repository_file_returns_empty_list(tmpdir):
    repo = RecipeRepository(tmpdir)
    assert repo.list() == []


def test_entries_from_json_show_up_in_list(tmpdir):
    today = date.today()
    repo = RecipeRepository(tmpdir)
    data = [
        {"date": str(today), "data": {"title": "test"}},
    ]
    with repo.path.open("w") as f:
        f.write(json.dumps(data))

    [recipe] = repo.list()
    assert recipe.date == today
    assert recipe.data == {"title": "test"}


def test_add_recipe_to_repository(tmpdir):
    repo = RecipeRepository(tmpdir)
    data = {"_id": {"$oid": 1, "title": "test"}}
    recipe = RecipeInDb(date=date.today(), data=data)
    repo.add(recipe)

    [recipe_from_repo] = repo.list()
    assert recipe_from_repo == recipe

    repo.add(recipe)
    assert len(repo.list()) == 1


def test_add_recipe_list_to_repository(tmpdir):
    repo = RecipeRepository(tmpdir)
    data1 = {"_id": {"$oid": 1, "title": "test"}}
    recipe1 = RecipeInDb(date=date.today(), data=data1)
    data2 = {"_id": {"$oid": 2, "title": "test"}}
    recipe2 = RecipeInDb(date=date.today(), data=data2)
    recipes = [recipe1, recipe2]
    repo.add_list(recipes)
    assert len(repo.list()) == 2


def test_needs_to_be_synced(tmpdir):
    repo = RecipeRepository(tmpdir)
    today = date.today()
    assert repo.needs_to_be_synced(today)

    data = {"_id": {"$oid": 1, "title": "test"}}
    recipe = RecipeInDb(date=today, data=data)
    repo.add(recipe)
    assert not repo.needs_to_be_synced(today)



from kptncook.mealie import MealieApiClient, Recipe


def test_parse_empty_mealie_recipe_is_valid():
    mealie_recipe = {}
    assert Recipe.model_validate(mealie_recipe) == Recipe()


def test_parse_mealie_recipe_pydantic_update():
    """
    https://github.com/ephes/kptncook/issues/28
    This recipe raised a pydantic ValidationError after pydantic2 upgrade:
    ValidationError: 2 validation errors for Recipe
    recipe_yield
      Field required [type=missing, input_value={'id': '7fb151b5-7807-4c2... None,
    'lastMade': None}, input_type=dict]
        For further information visit https://errors.pydantic.dev/2.4/v/missing
    nutrition
      Field required [type=missing, input_value={'id': '7fb151b5-7807-4c2... None,
    'lastMade': None}, input_type=dict]
        For further information visit https://errors.pydantic.dev/2.4/v/missing
    """
    mealie_recipe = {
        "id": "9e1f19c2-a087-4834-8cbb-9722ac1e49f3",
        "userId": "fcd14f40-ecc8-4b78-9727-dc9597fc208c",
        "groupId": "c32d7dd5-75c7-4e33-9fb9-f31005b85865",
        "name": "Mediterrane Hühnchen-Quinoa-Bowl",
        "slug": "mediterrane-huhnchen-quinoa-bowl",
        "image": None,
        "recipeYield": "1 Portionen",
        "totalTime": None,
        "prepTime": "30",
        "cookTime": "0",
        "performTime": None,
        "description": "",
        "recipeCategory": [],
        "tags": [
            {
                "id": "d3709045-6618-4faf-bb2d-f9ea6435c64b",
                "name": "kptncook",
                "slug": "kptncook",
            }
        ],
        "tools": [],
        "rating": None,
        "orgURL": None,
        "dateAdded": None,
        "dateUpdated": "2022-12-13T18:21:22.189956",
        "createdAt": None,
        "updateAt": None,
        "lastMade": None,
    }
    recipe = Recipe.model_validate(mealie_recipe)
    assert recipe.name == mealie_recipe["name"]


def test_skip_validation_errors():
    invalid_recipes = [
        {"id": "123", "nutrition": "not a nutrition dict"},
    ]
    assert MealieApiClient.validate_recipes(invalid_recipes) == []



import re
from datetime import date
from time import time
from typing import Literal
from urllib.parse import urljoin

import httpx

from .config import settings
from .repositories import RecipeInDb


def ids_to_payload(ids: list[tuple[str, str]]) -> list[dict]:
    """
    Convert a list of (type, id) tuples to a list of dicts that
    can be used as payload for the kptncook api.
    """
    payload = []
    for id_type, id_value in ids:
        if id_type == "oid":
            payload.append({"identifier": id_value})
        elif id_type == "uid":
            payload.append({"uid": id_value})
    return payload


class KptnCookClient:
    """
    Client for the kptncook api.
    """

    def __init__(
        self, base_url=settings.kptncook_api_url, api_key=settings.kptncook_api_key
    ):
        self.base_url = str(base_url)
        self.headers = {
            "content-type": "application/json",
            "Accept": "application/vnd.kptncook.mobile-v8+json",
            "User-Agent": "Platform/Android/12.0.1 App/7.10.1",
            "hasIngredients": "yes",
        }
        self.api_key = api_key
        if settings.kptncook_access_token is not None:
            self.headers["Token"] = settings.kptncook_access_token

    @property
    def logged_in(self):
        return "Token" in self.headers

    def to_url(self, path):
        return urljoin(self.base_url, path)

    def __getattr__(self, name):
        """
        Return proxy for httpx, joining base_url with path and
        providing authentication headers automatically.
        """

        def proxy(path, **kwargs):
            url = self.to_url(path)
            set_headers = kwargs.get("headers", {})
            kwargs["headers"] = set_headers | self.headers
            return getattr(httpx, name)(url, **kwargs)

        return proxy

    def list_today(self) -> list[RecipeInDb]:
        """
        Get all recipes for today from kptncook api.
        """
        time_str = str(time())
        response = self.get(f"recipes/de/{time_str}?kptnkey={self.api_key}")
        response.raise_for_status()
        recipes = []
        today = date.today()
        for data in response.json():
            recipes.append(RecipeInDb(date=today, data=data))
        return recipes

    def get_access_token(self, username: str, password: str) -> str:
        """
        Get access token for kptncook api.
        """
        response = self.post(
            "/login/userpass",
            json={"email": username, "password": password},
        )
        response.raise_for_status()
        token_data = response.json()
        return token_data["accessToken"]

    def list_favorites(self) -> list[str]:
        """
        Get list of favorite recipes.
        """
        response = self.get("/favorites")
        response.raise_for_status()
        return response.json()["favorites"]

    def get_by_ids(self, ids: list[tuple[str, str]]) -> list[RecipeInDb]:
        """
        Get recipes from list of ids.
        """
        payload = ids_to_payload(ids)
        # timeout disabled because saving more than 999 favorites didn't work for @brotkrume
        response = self.post(
            f"/recipes/search?kptnkey={self.api_key}", json=payload, timeout=None
        )
        response.raise_for_status()
        results = response.json()
        if results is None:
            results = []
        results = [RecipeInDb(date=date.today(), data=data) for data in results]
        return results


def looks_like_uid(token: str) -> bool:
    correct_len = len(token) == 8
    is_alnum = token.isalnum()
    return correct_len and is_alnum


def looks_like_oid(token: str) -> bool:
    correct_len = len(token) == 24
    is_hex = token.isalnum()
    return correct_len and is_hex


def parse_id(text: str) -> tuple[Literal["oid", "uid"], str] | None:
    """
    Parse recipe id (uid or uid) from url/text.
    """
    try:
        uid = next(part for part in re.split(r"/|\?", text) if looks_like_uid(part))
        return "uid", uid
    except StopIteration:
        pass
    try:
        oid = next(part for part in re.split(r" |,|/", text) if looks_like_oid(part))
        return "oid", oid
    except StopIteration:
        pass
    return None



"""
kptncook is a little command line utility to download
new recipes.
"""

import sys
from datetime import date
from typing import Optional

import httpx
import typer
from rich import print as rprint
from rich.pretty import pprint
from rich.prompt import Prompt

from .api import KptnCookClient, parse_id
from .config import settings
from .mealie import MealieApiClient, kptncook_to_mealie
from .models import Recipe
from .paprika import PaprikaExporter
from .repositories import RecipeRepository

__all__ = [
    "list_kptncook_today",
    "save_todays_recipes",
    "sync_with_mealie",
    "sync",
    "backup_kptncook_favorites",
    "get_kptncook_access_token",
    "list_recipes",
    "search_kptncook_recipe_by_id",
    "export_recipes_to_paprika",
]

__version__ = "0.0.18"
cli = typer.Typer()


@cli.command(name="kptncook-today")
def list_kptncook_today():
    """
    List all recipes for today from the kptncook site.
    """
    client = KptnCookClient()
    all_recipes = client.list_today()
    for recipe in all_recipes:
        pprint(recipe)


@cli.command(name="save-todays-recipes")
def save_todays_recipes():
    """
    Save recipes for today from kptncook site.
    """
    fs_repo = RecipeRepository(settings.root)
    if fs_repo.needs_to_be_synced(date.today()):
        client = KptnCookClient()
        fs_repo.add_list(client.list_today())


def get_mealie_client() -> MealieApiClient:
    client = MealieApiClient(settings.mealie_url)
    client.login(settings.mealie_username, settings.mealie_password)
    return client


def get_kptncook_recipes_from_mealie(client):
    recipes = client.get_all_recipes()
    recipes_with_details = []
    for recipe in recipes:
        recipes_with_details.append(client.get_via_slug(recipe.slug))
    kptncook_recipes = [
        r for r in recipes_with_details if r.extras.get("source") == "kptncook"
    ]
    return kptncook_recipes


def get_kptncook_recipes_from_repository() -> list[Recipe]:
    fs_repo = RecipeRepository(settings.root)
    recipes = []
    for repo_recipe in fs_repo.list():
        recipes.append(Recipe.model_validate(repo_recipe.data))
        # try:
        #     recipes.append(Recipe.model_validate(repo_recipe.data))
        # except Exception as e:
        #     print(f"Could not parse recipe {repo_recipe.id}: {e}")
        #     for ingredient in repo_recipe.data.get("ingredients"):
        #         uncountable_title = ingredient["ingredient"].get("uncountableTitle")
        #         if uncountable_title is None:
        #             print("ingredient: ", ingredient["ingredient"])
        #     return []
    return recipes


def get_recipe_from_repository_by_oid(oid: str) -> list[Recipe]:
    """
    get one single recipe from local repository
    :param oid: oid of recipe
    :return: list
    """
    recipes = get_kptncook_recipes_from_repository()
    return [recipe for num, recipe in enumerate(recipes) if recipe.id.oid == oid]


@cli.command(name="sync-with-mealie")
def sync_with_mealie():
    """
    Sync locally saved recipes with mealie.
    """
    try:
        client = get_mealie_client()
    except Exception as e:
        rprint(f"Could not login to mealie: {e}")
        sys.exit(1)
    kptncook_recipes_from_mealie = get_kptncook_recipes_from_mealie(client)
    recipes = get_kptncook_recipes_from_repository()
    kptncook_recipes_from_repository = [kptncook_to_mealie(r) for r in recipes]
    ids_in_mealie = {r.extras["kptncook_id"] for r in kptncook_recipes_from_mealie}
    ids_from_api = {r.extras["kptncook_id"] for r in kptncook_recipes_from_repository}
    ids_to_add = ids_from_api - ids_in_mealie
    recipes_to_add = []
    for recipe in kptncook_recipes_from_repository:
        if recipe.extras.get("kptncook_id") in ids_to_add:
            recipes_to_add.append(recipe)
    created_slugs = []
    for recipe in recipes_to_add:
        try:
            created = client.create_recipe(recipe)
            created_slugs.append(created.slug)
        except httpx.HTTPStatusError as e:
            if (
                e.response.json().get("detail", {}).get("message")
                == "Recipe already exists"
            ):
                continue
    rprint(f"Created {len(created_slugs)} recipes")


@cli.command(name="sync")
def sync():
    """
    Fetch recipes for today from api, save them to disk and sync with mealie
    afterwards.
    """
    save_todays_recipes()
    sync_with_mealie()


@cli.command(name="backup-favorites")
def backup_kptncook_favorites():
    """
    Store kptncook favorites in local repository.
    """
    if settings.kptncook_access_token is None:
        rprint("Please set KPTNCOOK_ACCESS_TOKEN in your environment or .env file")
        sys.exit(1)
    client = KptnCookClient()
    favorites = client.list_favorites()
    rprint(f"Found {len(favorites)} favorites")
    ids = [("oid", oid["identifier"]) for oid in favorites]
    recipes = client.get_by_ids(ids)
    if len(recipes) == 0:
        rprint("Could not find any favorites")
        sys.exit(1)

    fs_repo = RecipeRepository(settings.root)
    fs_repo.add_list(recipes)
    rprint(f"Added {len(recipes)} recipes to local repository")


@cli.command(name="kptncook-access-token")
def get_kptncook_access_token():
    """
    Get access token for kptncook.
    """
    username = Prompt.ask("Enter your kptncook email address")
    password = Prompt.ask("Enter your kptncook password", password=True)
    client = KptnCookClient()
    access_token = client.get_access_token(username, password)
    rprint("your access token: ", access_token)


@cli.command(name="list-recipes")
def list_recipes():
    """
    List all locally saved recipes.
    """
    recipes = get_kptncook_recipes_from_repository()
    for num, recipe in enumerate(recipes):
        rprint(num, recipe.localized_title.de, recipe.id.oid)


@cli.command(name="search-by-id")
def search_kptncook_recipe_by_id(id_: str):
    """
    Search for a recipe by id in kptncook api, id can be a sharing
    url or an oid for example, and add it to the local repository.
    """
    if id_.startswith(
        "https://share.kptncook.com/"
    ):  # sharing url -> use redirect location
        r = httpx.get(id_)
        if r.status_code not in (301, 302):
            rprint("Could not get redirect location")
            sys.exit(1)
        id_ = r.headers["location"]
    parsed = parse_id(id_)
    if parsed is None:
        rprint("Could not parse id")
        sys.exit(1)
    id_type, id_value = parsed
    rprint(id_type, id_value)
    client = KptnCookClient()
    recipes = client.get_by_ids([(id_type, id_value)])
    if len(recipes) == 0:
        rprint("Could not find recipe")
        sys.exit(1)
    recipe = recipes[0]
    fs_repo = RecipeRepository(settings.root)
    fs_repo.add_list([recipe])
    rprint(f"Added recipe {id_type} {id_value} to local repository")


# Optional needed by typer, standalone to trick pyupgrade to not change it
OptionalId = Optional[str]


@cli.command(name="export-recipes-to-paprika")
def export_recipes_to_paprika(_id: OptionalId = typer.Argument(None)):
    """
    Export one recipe or all recipes to Paprika app

    Example usage 1:  kptncook  export-recipes-to-paprika 635a68635100007500061cd7
    Example usage 2:  kptncook  export-recipes-to-paprika
    """
    if _id:
        recipes = get_recipe_by_id(_id)
    else:
        recipes = get_kptncook_recipes_from_repository()
    exporter = PaprikaExporter()
    filename = exporter.export(recipes=recipes)
    rprint(
        "\n The data was exported to '%s'. Open the export file with the Paprika App.\n"
        % filename
    )


def get_recipe_by_id(_id: str):
    parsed = parse_id(_id)
    if parsed is None:
        rprint("Could not parse id")
        sys.exit(1)
    # we can expect always an oid here - correct?
    id_type, id_value = parsed
    found_recipes = get_recipe_from_repository_by_oid(oid=id_value)
    if len(found_recipes) == 0:
        rprint("Recipe not found.")
        sys.exit(1)
    if len(found_recipes) > 1:
        rprint("More than one recipe found with that ID.")
        sys.exit(1)
    return found_recipes


if __name__ == "__main__":
    cli()



import os

import httpx
import pytest

from kptncook.models import Recipe
from kptncook.paprika import GeneratedData, PaprikaExporter


def test_asciify_string():
    p = PaprikaExporter()
    assert (
        p.asciify_string("Süßkartoffeln mit Taboulé & Dip")
        == "Susskartoffeln_mit_Taboule___Dip"
    )
    assert p.asciify_string("Ölige_Ähren") == "Olige_Ahren"


def test_get_cover_img_as_base64_string(full_recipe, mocker):
    p = PaprikaExporter()
    recipe = Recipe.model_validate(full_recipe)
    mocker.patch(
        "kptncook.paprika.httpx.get",
        return_value=mocker.Mock(content=b"foobar", status_code=200),
    )
    cover_info = p.get_cover_img_as_base64_string(recipe=recipe)
    assert isinstance(cover_info, tuple) is True
    assert len(cover_info) == 2

    # no images available for some reason
    recipe.image_list = list()
    with pytest.raises(ValueError):
        p.get_cover_img_as_base64_string(recipe=recipe)


def test_get_cover_img_as_base64_string_can_handle_404(full_recipe, mocker):
    p = PaprikaExporter()
    recipe = Recipe.model_validate(full_recipe)
    mocker.patch(
        "kptncook.paprika.httpx.get",
        return_value=mocker.Mock(content=b"foobar", status_code=404),
    )
    # hm, looks weird, but works.
    m = mocker.patch("kptncook.paprika.httpx.get")
    mock_response = mocker.Mock()
    mock_response.raise_for_status = mocker.Mock(
        side_effect=httpx.HTTPStatusError(
            message="404 File not found", response=mock_response, request=mocker.Mock()
        )
    )
    m.return_value = mock_response
    assert p.get_cover_img_as_base64_string(recipe=recipe) == (None, None)


def test_export_single_recipe(full_recipe, mocker):
    p = PaprikaExporter()
    recipe = Recipe.model_validate(full_recipe)
    mocker.patch(
        "kptncook.paprika.httpx.get",
        return_value=mocker.Mock(content=b"foobar", status_code=200),
    )
    p.export(recipes=[recipe])
    expected_file = (
        "Uberbackene_Muschelnudeln_mit_Lachs___Senf_Dill_Sauce.paprikarecipes"
    )
    assert os.path.isfile(expected_file) is True
    if os.path.isfile(expected_file):
        os.unlink(expected_file)


def test_export_all_recipes(full_recipe, minimal, mocker):
    p = PaprikaExporter()
    recipe1 = Recipe.model_validate(full_recipe)
    recipe2 = Recipe.model_validate(minimal)
    mocker.patch(
        "kptncook.paprika.httpx.get",
        return_value=mocker.Mock(content=b"foobar", status_code=200),
    )
    p.export(recipes=[recipe1, recipe2])
    expected_file = "allrecipes.paprikarecipes"
    assert os.path.isfile(expected_file) is True
    if os.path.isfile(expected_file):
        os.unlink(expected_file)


def test_get_cover(minimal):
    p = PaprikaExporter()
    recipe = Recipe.model_validate(minimal)
    assert p.get_cover(image_list=list()) is None

    cover = p.get_cover(image_list=recipe.image_list)
    assert cover.name == "REZ_1837_Cover.jpg"

    with pytest.raises(ValueError):
        p.get_cover(image_list=None)

    with pytest.raises(ValueError):
        p.get_cover(image_list=dict())


def test_render(minimal, mocker):
    # happy path
    recipe = Recipe.model_validate(minimal)
    paprika_exporter = PaprikaExporter()
    generated = GeneratedData("", "", "", "")
    mocker.patch.object(
        paprika_exporter, "get_generated_data", return_value=generated, autospec=True
    )
    export_data = paprika_exporter.get_export_data(recipes=[recipe])
    [actual] = list(export_data.values())
    expected = (
        "{\n"
        '   "uid":"5e5390e2740000cdf1381c64",\n'
        '   "name":"Minimal Recipe",\n'
        '   "directions": "Alles parat?\\n",\n'
        '   "servings":"2",\n'
        '   "rating":0,\n'
        '   "difficulty":"",\n'
        '   "ingredients":"",\n'
        '   "notes":"",\n'
        '   "created":"",\n'
        '   "image_url":null,\n'
        '   "cook_time":"",\n'
        '   "prep_time":"20",\n'
        '   "source":"Kptncook",\n'
        '   "source_url":"",\n'
        '   "hash" : "",\n'
        '   "photo_hash":null,\n'
        '   "photos":[],\n'
        '   "photo": "",\n'
        '   "nutritional_info":"calories: 100\\nprotein: 30\\nfat: 10\\ncarbohydrate: '
        '20\\n",\n'
        '   "photo_data":"",\n'
        '   "photo_large":null,\n'
        '   "categories":["Kptncook"]\n'
        "}"
    )
    assert actual == expected



from kptncook.models import IngredientDetails, Recipe, to_camel


def test_to_camel():
    assert to_camel("localized_title") == "localizedTitle"
    assert to_camel("foo") == "foo"


def test_parse_recipe_id(minimal):
    recipe = Recipe.model_validate(minimal)
    assert recipe.id.oid == minimal["_id"]["$oid"]


def test_parse_image_url(minimal):
    recipe = Recipe.model_validate(minimal)
    image_url = recipe.get_image_url("foobar")
    assert image_url is not None
    assert "foobar" in image_url

    recipe = recipe.model_copy()
    recipe.image_list = []
    assert recipe.get_image_url("foobar") is None


def test_parse_full_recipe(full_recipe):
    recipe = Recipe.model_validate(full_recipe)
    assert recipe is not None
    # from rich.pretty import pprint

    # pprint(full_recipe)
    # assert False


def test_ingredient_details_without_uncountable_title():
    ingredient_details = {
        "typ": "ingredient",
        "localizedTitle": {"de": "Zucker", "en": "sugar"},
        "numberTitle": {"de": "Zucker", "en": "sugar"},
        "uncountableTitle": None,
        "category": "baking",
    }
    ingredient_details = IngredientDetails(**ingredient_details)
    assert ingredient_details.uncountable_title == ingredient_details.number_title



"""
Base settings for kptncook.
"""
import sys
from pathlib import Path

from pydantic import AnyHttpUrl, DirectoryPath, Field, ValidationError, field_validator
from pydantic_settings import BaseSettings, SettingsConfigDict
from rich import print as rprint


class Settings(BaseSettings):
    model_config = SettingsConfigDict(
        env_file=(Path.home() / ".kptncook" / ".env"), extra="ignore"
    )
    root: DirectoryPath = Field(Path.home() / ".kptncook", env="KPTNCOOK_HOME")
    kptncook_api_key: str
    kptncook_access_token: str | None = None
    kptncook_api_url: AnyHttpUrl = AnyHttpUrl("https://mobile.kptncook.com")
    mealie_url: AnyHttpUrl = AnyHttpUrl("http://localhost:9000/api")
    mealie_username: str
    mealie_password: str

    @field_validator("root", mode="before")
    def root_must_exist(cls, path: Path) -> Path:
        path.mkdir(parents=True, exist_ok=True)
        return path


try:
    settings = Settings()  # type: ignore
except ValidationError as e:
    rprint("validation error: ", e)
    sys.exit(1)



import pytest

from kptncook.api import parse_id


@pytest.mark.parametrize(
    "test_input,expected",
    [
        ("", None),
        ("asdf", None),
        (
            (
                "https://mobile.kptncook.com/recipe/pinterest/"
                "Rote-Pasta-mit-Salbei-Zitronen-Butter-&-karamell"
                "isierten-Waln%C3%BCssen/19e2eda2?_branch_match_id"
                "=866187386318966341&utm_source=SMS&utm_medium=sharing"
                "&_branch_referrer=H4sIAAAAAAAAA8soKSkottLXL85ILMrMS9fL"
                "LijJS87Pz9ZLzs%2FVN3b2SY4sTzatyk8CACm3OLEoAAAA"
            ),
            ("uid", "19e2eda2"),
        ),
        (
            (
                "https://mobile.kptncook.com/recipe/pinterest/"
                "Rote-Pasta-mit-Salbei-Zitronen-Butter-&-karamell"
                "isierten-Waln%C3%BCssen/19e2eda2?_branch_match_id"
                "=866187386318966341&utm_source=SMS&utm_medium=sharing"
                "&_branch_referrer=H4sIAAAAAAAAA8soKSkottLXL85ILMrMS9fL"
                "LijJS87Pz9ZLzs%2FVN3b2SY4sTzatyk8CACm3OLEoAAAA"
            ),
            ("uid", "19e2eda2"),
        ),
        ("19e2eda2", ("uid", "19e2eda2")),
        ("5aa2cbb028000052091b5c6c", ("oid", "5aa2cbb028000052091b5c6c")),
    ],
)
def test_parse_id(test_input, expected):
    print(test_input, expected)
    assert parse_id(test_input) == expected



"""
Repositories to store recipes.

Atm only uses json. But this could also be a sqlite or a
remote api, maybe mealie... hmm.
"""
import shutil
from datetime import date
from pathlib import Path
from typing import List  # noqa F401

from pydantic import BaseModel, RootModel


class RecipeInDb(BaseModel):
    date: date
    data: dict

    @property
    def id(self):
        return self.data["_id"]["$oid"]


class RecipeListInDb(RootModel):
    root: list[RecipeInDb]

    def __iter__(self):
        return iter(self.root)

    def __getitem__(self, item):
        return self.root[item]


class RecipeRepository:
    name: str = "kptncook.json"

    def __init__(self, base_dir: Path):
        self.base_dir = base_dir

    @property
    def path(self) -> Path:
        return self.base_dir / self.name

    @property
    def backup_path(self) -> Path:
        return self.base_dir / f"{self.name}.backup"

    def create_backup(self):
        if self.path.exists():
            shutil.copyfile(self.path, self.backup_path)

    def _write_models(self, locked):
        self.create_backup()
        try:
            self.path.parent.mkdir(exist_ok=True)
        except AttributeError:
            # LocalPath in tests
            pass
        models = RecipeListInDb.model_validate(locked.values())
        with self.path.open("w") as f:
            f.write(models.model_dump_json())

    def _fetch_all(self):
        """
        Fetch dict of pydantic models from json in self.path
        """
        try:
            if not self.path.exists():
                return []
            with self.path.open("r") as f:
                recipes_in_db = RecipeListInDb.model_validate_json(f.read())
            return recipes_in_db
        except FileNotFoundError:
            return []

    def list_by_id(self):
        by_id = {}
        for recipe in self.list():
            by_id[recipe.id] = recipe
        return by_id

    def needs_to_be_synced(self, _date: date):
        """
        Return True if there are no recipes for date.
        """
        return not any(recipe.date == _date for recipe in self.list())

    def add(self, recipe: RecipeInDb):
        locked = self.list_by_id()
        locked[recipe.id] = recipe
        self._write_models(locked)

    def add_list(self, recipes: list[RecipeInDb]):
        locked = self.list_by_id()
        for recipe in recipes:
            locked[recipe.id] = recipe
        self._write_models(locked)

    def list(self):
        return list(self._fetch_all())
